# Copyright (C) 2019 Majormode.  All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import collections
import os

from majormode.mercurius.constant.place import SearchIntent
from majormode.mercurius.model.place import AddressComponentName
from majormode.perseus.constant.obj import ObjectStatus
from majormode.perseus.constant.privacy import Visibility
from majormode.perseus.model.geolocation import GeoPoint
from majormode.perseus.model.locale import DEFAULT_LOCALE
from majormode.perseus.model.obj import Object
from majormode.perseus.service.base_rdbms_service import BaseRdbmsService
from majormode.perseus.service.area.area_service import AreaService
from majormode.perseus.utils import cast
from majormode.perseus.utils import string_util

import settings


class PlaceService(BaseRdbmsService):
    # Name of the bucket on a local or remote Content Delivery Network
    # (CDN) where the thumbnail versions of place logo pictures are stored
    # into.
    CDN_BUCKET_NAME_LOGO = 'logo'

    # Name of the bucket on a local or remote Content Delivery Network
    # (CDN) where the thumbnail versions of photos are stored into.
    CDN_BUCKET_NAME_PHOTO = 'photo'

    # Default distance in meters of the radius of search for places nearby
    # a given location.
    DEFAULT_SEARCH_RADIUS = 20000

    # Maximal distance in meters of the radius of search for places nearby
    # a given location.
    MAXIMAL_SEARCH_RADIUS = 40000

    # PlaceNotification = Enum(
    # )

    def __add_place_address(
            self,
            place_id,
            address,
            account_id=None,
            connection=None,
            locale=DEFAULT_LOCALE):
        """
        Add a list of address information to a place.


        :param place_id: Identification of a place.

        :param address: A list of `Contact` objects.

        :param account_id: Identification of the account of the user who adds
            this contact information to the place.

        :param connection: a `RdbmsConnection` instance to be used supporting
            the Python clause `with ...:`.

        :param locale: An object `Locale` defining the language which the
            textual information of the address component is written in.
        """
        with self.acquire_rdbms_connection(auto_commit=True, connection=connection) as connection:
            connection.execute(
                """
                INSERT INTO place_address(
                    place_id,
                    account_id,
                    locale,
                    property_name,
                    property_value)
                  VALUES 
                    %[values]s
                """,
                {
                    'values': [
                        (place_id,
                         account_id,
                         locale or DEFAULT_LOCALE,
                         property_name,
                         property_value)
                        for (property_name, property_value) in address]
                })

    def __add_place_contacts(
            self,
            place_id,
            contacts,
            account_id=None,
            connection=None):
        """
        Add a list of contact information to a place.


        :param place_id: Identification of a place.

        :param contacts: A list of `Contact` objects.

        :param account_id: Identification of the account of the user who adds
            this contact information to the place.

        :param connection: a `RdbmsConnection` instance to be used supporting
            the Python clause `with ...:`.
        """
        with self.acquire_rdbms_connection(auto_commit=True, connection=connection) as connection:
            connection.execute(
                """
                INSERT INTO 
                    place_contact(
                        place_id,
                        account_id,
                        property_name,
                        property_value,
                        is_primary,
                        is_verified,
                        visibility)
                  VALUES 
                    %[values]s
                    """,
                {
                    'values': [
                        (place_id,
                         account_id,
                         contact.name,
                         contact.value,
                         contact.is_primary,
                         contact.is_verified,
                         contact.visiblity)
                        for contact in contacts
                    ]
                })

    @classmethod
    def build_picture_url(cls, picture_id):
        """
        Return the Uniform Resource Locator of a place's picture (logo).


        :param picture_id: Identification of the logo of a place.


        :return: A string representing the Uniform Resource Locator of the
            picture.
        """
        return picture_id and os.path.join(
            settings.CDN_URL_HOSTNAME,
            cls.CDN_BUCKET_NAME_LOGO,
            str(picture_id))

    def get_place_address(
            self,
            place_id,
            account_id=None,
            connection=None,
            locale=None):
        """
        Return the components of the postal address of a place.


        :param place_id: Identification of the place to return information.

        :param account_id: Identification of the account of behalf of whom
            the function is called.  If defined, the function returns
            customized information that this user may have defined for this
            place.

        :param connection: A `RdbmsConnection` instance to be used supporting
            the Python clause `with ...:`.

        :param locale: An object `Locale` to return textual information of
            the place.


        :return: A dictionary where the key corresponds to an item of the
            enumeration `AddressComponentName` and the value corresponds to
            the address component's value.
        """
        with self.acquire_rdbms_connection(auto_commit=False, connection=connection) as connection:
            cursor = connection.execute(
                """
                SELECT 
                    account_id,
                    property_name,
                    property_value,
                    locale
                  FROM (
                    SELECT
                        account_id,
                        property_name,
                        property_value,
                        locale,
                        row_number() OVER (
                            PARTITION BY 
                                place_id,
                                property_name
                            ORDER BY 
                                compare_locale(locale, %(locale)s) DESC,
                                account_id NULLS LAST) AS position
                      FROM place_address
                      WHERE place_id = %(place_id)s
                        AND ((%(account_id)s IS NULL AND account_id IS NULL) OR
                             (%(account_id)s IS NOT NULL 
                              AND (account_id IS NULL OR account_id = %(account_id)s)))
                        AND (compare_locale(locale, %(locale)s) > 0 
                             OR compare_locale(locale, %(DEFAULT_LOCALE)s) > 0)) AS foo
                  WHERE position = 1
                """,
                {
                    'DEFAULT_LOCALE': DEFAULT_LOCALE,
                    'account_id': account_id,
                    'locale': locale or DEFAULT_LOCALE,
                    'place_id': place_id
                })

            # Build the dictionary of the address components registered for this
            # place.
            #
            # @todo: Do we need to also return the locale and the identification of the account?
            address = dict([
                (component.property_name, component.property_value)
                for component in [
                    row.get_object({
                        'account_id': cast.string_to_uuid,
                        'locale': cast.string_to_locale,
                        'property_name': AddressComponentName})
                    for row in cursor.fetch_all()
                ]
            ])

            return address

    def get_place_contacts(
            self,
            place_id,
            account_id=None,
            connection=None):
        """

        :param place_id:
        :param account_id:
        :param connection:
        :return:
        """
        with self.acquire_rdbms_connection(auto_commit=False, connection=connection) as connection:
            cursor = connection.execute(
                """
                SELECT 
                    *
                  FROM (
                    SELECT
                        property_name,
                        property_value,
                        visibility,
                        is_primary,
                        is_verified,
                        row_number() OVER (
                            PARTITION BY 
                                place_id, 
                                property_name
                            ORDER BY 
                                account_id NULLS LAST) AS position
                      FROM place_contact
                      WHERE place_id = %(place_ids)s
                        AND ((%(account_id)s IS NULL AND account_id IS NULL) OR
                             (%(account_id)s IS NOT NULL AND (account_id IS NULL OR account_id = %(account_id)s)))) AS foo
                    WHERE position = 1
                """,
                {
                    'account_id': account_id,
                    'place_id': place_id
                })

            contacts = [
                row.get_object({'visibility': Visibility})
                for row in cursor.fetch_all()
            ]

            return contacts

    def __get_place_photos(
            self,
            place_id,
            account_id=None,
            connection=None,
            limit=None,
            offset=0):
        """
        Return a list of photos related to a place.


        :param place_id: Identification of a place.

        :param account_id: Identification of the account of a user on behalf
            of whom this function is called.

        :param connection: A `RdbmsConnection` instance to be used supporting
            the Python clause `with ...:`.

        :param limit: Constrain the number of photos that are returned to the
            specified number.

        :param offset: Require to skip that many photos before beginning to
            return them.  If both `limit` and `offset` are specified, then
            `offset` photos are skipped before starting to count the `limit`
            photos that are returned.  The default value is `0`.


        :return: A list of objects containing the following attributes:

            * `account_id`: Identification of the account of the user who posted
              this photo.

            * `capture_time` (optional): Time when this photo has been captured.

            * `creation_time` (required): Time when this photo has been registered
              to the platform.

            * `object_status` (required): Current status of this photo.

            * `photo_id` (required): Identification of a photo.

            * `team_id` (optional): Identification of the organization on behalf of
              which this photo has been posted.

            * `visibility` (required): Visibility of this photo to other users.
        """
        with self.acquire_rdbms_connection(auto_commit=False, connection=connection) as connection:
            cursor = connection.execute(
                """
                SELECT 
                    photo_id,
                    account_id,
                    team_id,
                    capture_time,
                    object_status,
                    creation_time
                  FROM 
                    place_photo
                  WHERE
                    place_id = %(place_id)s
                    AND (object_status = %(OBJECT_STATUS_ENABLED)s
                         OR (object_status IN (%(OBJECT_STATUS_PENDING)s, %(OBJECT_STATUS_DISABLED)s)  
                             AND (account_id = %(account_id)s 
                                  OR (team_id IS NOT NULL AND team_is_administrator(account_id, team_id)))))
                    AND (visibility = %(VISIBILITY_PUBLIC)s 
                         OR (visibility = %(VISIBILITY_PRIVATE)s AND account_id = %(account_id)s)
                         OR (visibility = %(VISIBILITY_TEAM)s
                             AND (account_id = %(account_id)s OR team_is_member(account_id, team_id))))
                  ORDER BY
                    creation_time DESC
                  LIMIT %(limit)s
                  OFFSET %(offset)s
                """,
                {
                    'OBJECT_STATUS_DISABLED': ObjectStatus.disabled,
                    'OBJECT_STATUS_ENABLED': ObjectStatus.enabled,
                    'OBJECT_STATUS_PENDING': ObjectStatus.pending,
                    'VISIBILITY_PRIVATE': Visibility.private,
                    'VISIBILITY_PUBLIC': Visibility.public,
                    'VISIBILITY_TEAM': Visibility.team,
                    'account_id': account_id,
                    'limit': min(limit, self.MAXIMUM_LIMIT) or self.MAXIMUM_LIMIT,
                    'offset': offset,
                    'place_id': place_id
                })

            photos = [
                row.get_object({
                    'account_id': cast.string_to_uuid,
                    'capture_time': cast.string_to_timestamp,
                    'creation_time': cast.string_to_timestamp,
                    'object_status': ObjectStatus,
                    'photo_id': cast.string_to_uuid,
                    'team_id': cast.string_to_uuid,
                    'visibility': Visibility
                })
                for row in cursor.fetch_all()
            ]

            return photos

    def get_places_address(
            self,
            place_ids,
            account_id=None,
            connection=None,
            locale=None):
        """
        Return the address information of a list of places.


        :param place_ids: List of identifications of the places to return the
            address information.

        :param account_id: Identification of the account of behalf of whom
            the function is called.  If defined, the function returns
            customized information that this user may have defined for this
            place.

        :param connection: An object ``RdbmsConnection`` supporting the Python
            clause ``with ...:``, or ``None``.

        :param locale: An object `Locale` to return textual information of the
            place.


        :return: A dictionary of address information where a key corresponds
            to the identification of a place, and the value is a dictionary of
            address components.
        """
        with self.acquire_rdbms_connection(auto_commit=False, connection=connection) as connection:
            cursor = connection.execute(
                """
                SELECT *
                  FROM (
                    SELECT 
                        place_id,
                        account_id,
                        property_name,
                        property_value,
                        locale,
                        row_number() OVER (
                          PARTITION BY 
                            place_id,
                            property_name
                          ORDER BY
                            compare_locale(locale, %(locale)s) DESC,
                            account_id NULLS LAST) AS position
                      FROM
                        place_address
                      WHERE
                        place_id IN (%[place_ids]s)
                        AND ((%(account_id)s IS NULL AND account_id IS NULL) 
                             OR (%(account_id)s IS NOT NULL 
                                 AND (account_id IS NULL OR account_id = %(account_id)s)))
                        AND (compare_locale(locale, %(locale)s) > 0 
                             OR compare_locale(locale, %(DEFAULT_LOCALE)s) > 0)) AS foo
                    WHERE
                      position = 1
                """,
                {
                    'DEFAULT_LOCALE': DEFAULT_LOCALE,
                    'account_id': account_id,
                    'locale': locale and DEFAULT_LOCALE,
                    'place_ids': place_ids
                })

            addresses = [
                row.get_object({'place_id': cast.string_to_uuid})
                for row in cursor.fetch_all()
            ]

            places_address = collections.defaultdict(dict)
            for address in addresses:
                places_address[address.place_id][address.property_name] = address.property_value

            return places_address

    def get_places_contacts(
            self,
            place_ids,
            account_id=None,
            connection=None):
        """
        Return the contact information of a list of places.


        :param place_ids: List of identifications of the places to return the
            contact information.

        :param account_id: Identification of the account of behalf of whom
            the function is called.  If defined, the function returns
            customized information that this user may have defined for this
            place.

        :param connection: An object ``RdbmsConnection`` supporting the Python
            clause ``with ...:``, or ``None``.


        :return: A dictionary of contact information where a key corresponds
            to the identification of a place, and the value is a dictionary of
            contact information.
        """
        with self.acquire_rdbms_connection(auto_commit=False, connection=connection) as connection:
            cursor = connection.execute(
                """
                SELECT *
                  FROM (
                    SELECT 
                        place_id,
                        property_name,
                        property_value,
                        is_primary,
                        is_verified,
                        row_number() OVER (
                          PARTITION BY
                            place_id, 
                            property_name
                          ORDER BY
                            account_id NULLS LAST) AS position
                      FROM
                        place_contact
                      WHERE
                        place_id IN (%[place_ids]s)
                        AND ((%(account_id)s IS NULL AND account_id IS NULL) 
                             OR (%(account_id)s IS NOT NULL 
                                 AND (account_id IS NULL OR account_id = %(account_id)s)))) AS foo
                  WHERE position = 1
                """,
                {
                    'account_id': account_id,
                    'place_ids': place_ids
                })

            contacts = [
                row.get_object({'place_id': cast.string_to_uuid})
                for row in cursor.fetch_all()
            ]

            places_contacts = collections.defaultdict(dict)
            for contact in contacts:
                places_contacts[contact.place_id][contact.property_name] = contact.property_value

            return places_contacts

    def __index_place(self, place_id, connection=None):
        """
        Index a place with its information.

        The function retrieves the address component `recipient_name` and
        latinizes each word to index the place with.


        :param place_id: Identification of the place to index.

        :param connection: An object `RdbmsConnection` supporting the Python
            clause `with ...:`.
        """
        with self.acquire_rdbms_connection(auto_commit=True, connection=connection) as connection:
            cursor = connection.execute(
                """
                SELECT 
                    property_value AS name
                  FROM 
                    place_address
                  WHERE 
                    place_id = %(place_id)s
                    AND property_name = %(ADDRESS_COMPONENT_RECIPIENT_NAME)s
                """,
                {
                    'ADDRESS_COMPONENT_RECIPIENT_NAME': AddressComponentName.recipient_name,
                    'place_id': place_id
                })

            keywords = string_util.string_to_keywords(' '.join([
                row.get_value('name')
                for row in cursor.fetch_all()]))

            if keywords:
                connection.execute(
                    """
                    DELETE FROM 
                        place_index 
                      WHERE 
                        place_id = %(place_id)s
                    """,
                    {
                        'place_id': place_id
                    })

                connection.execute(
                    """
                    INSERT INTO place_index(
                        place_id,
                        keyword)
                      VALUES %[values]s
                    """,
                    {
                        'values': [(place_id, keyword) for keyword in keywords]
                    })

    def add_place(
            self,
            app_id,
            account_id,
            boundaries=None,
            category=None,
            connection=None,
            contacts=None,
            address=None,
            is_address_edited=False,
            is_location_edited=False,
            locale=None,
            location=None):
        """
        Register a new place.


        :param app_id: Identification of the client application such as a Web,
            a desktop, or a mobile application, that accesses the service.

        :param account_id: Identification of the account of of the user who
            adds this place.

        :param boundaries: A collection of one or more polygons that delimit
            the topological space of the place.  All of the polygons are
            within the spatial reference system.

        :param category: A string representation, or an item of an enumeration,
            of a category qualifying this place.

        :param connection: a `RdbmsConnection` instance to be used supporting
            the Python clause `with ...:`.

        :param contacts: A list of `Contact` objects representing the contact
            information of the place.

        :param address: Postal address of places, composed of one or more
            address components, which textual information is written in the
            specified locale.  An address component is defined with a
            component name (an item of the enumeration `AddressComponentName`)
            and its corresponding value.

        :param is_address_edited: Indicate whether the user has manually
            edited the address, or whether a reverse geocoder has
            automatically provided this address.

        :param is_location_edited: Indicate whether the location has been
            manually edited by the user, more likely by dragging/dropping a
            marker on an electronic map, or whether the device of the user has
            detected the user's current location or a geocoder has converted
            the formatted address of this place to a location.

        :param locale: An object `Locale` defining the language which the
            textual information of the address component is written in.

        :param location: An object `GeoPoint` representing the geographical
            location of the place's location (e.g., main entrance).


        :return: An object containing the following attributes:

            * `place_id` (required): Identification of the place.

            * `update_time` (required): Time of the most recent modification of
              one or more attributes of this place.
        """
        with self.acquire_rdbms_connection(auto_commit=True, connection=connection) as connection:
            cursor = connection.execute(
                """
                INSERT INTO place(
                    account_id,
                    category,
                    location,
                    boundaries,
                    is_address_edited,
                    is_location_edited)
                  VALUES 
                    (%(account_id)s,
                     %(category)s,
                     ST_SetSRID(ST_MakePoint(%(longitude)s, %(latitude)s, %(altitude)s), 4326),
                     ST_MakePolygon(ST_GeomFromText(%(boundaries)s, 4326)),                     
                     %(is_address_edited)s,
                     %(is_location_edited)s)
                  RETURNING 
                    place_id,
                    update_time
                """,
                {
                    'account_id': account_id,
                    'altitude': location and location.altitude,
                    'boundaries': boundaries and 'LINESTRING(%s)' % ','.join([
                        f'{longitude} {latitude} {altitude}'
                        for longitude, latitude, altitude in boundaries]),
                    'category': category,  # @todo: `category.strip().lower()` when category won't be a enumeration anymore
                    'is_address_edited': is_address_edited,
                    'is_location_edited': is_location_edited,
                    'latitude': location and location.latitude,
                    'longitude': location and location.longitude
                })

            place = cursor.fetch_one().get_object({
                'place_id': cast.string_to_uuid,
                'update_time': cast.string_to_timestamp
            })

            # Add the address information of the place, if any defined.
            if address:
                self.__add_place_addresss(
                    place.place_id,
                    address,
                    account_id=account_id,
                    connection=connection,
                    locale=locale)

            # Add the contact information of the place, if any defined.
            if contacts:
                self.__add_place_contacts(
                    place.place_id,
                    contacts,
                    account_id=account_id,
                    connection=connection)

            # Index the place so that it can be searchable providing a list of
            # keywords.
            self.__index_place(place.place_id, connection=connection)

            return place

    @classmethod
    def build_cover_photo_url(cls, picture_id):
        """
        Return the Uniform Resource Locator of the cover photos of a place.


        :param picture_id: Identification of the place's cover photo.


        :return: A string representing the Uniform Resource Locator of the
            place's cover photo.
        """
        return picture_id and os.path.join(
            settings.CDN_URL_HOSTNAME,
            cls.CDN_BUCKET_NAME_PHOTO,
            str(picture_id))

    @classmethod
    def build_picture_url(cls, picture_id):
        """
        Return the Uniform Resource Locator of a place's logo.


        :param picture_id: Identification of the place's logo.


        :return: A string representing the Uniform Resource Locator of the
            place's logo.
        """
        return picture_id and os.path.join(
            settings.CDN_URL_HOSTNAME,
            cls.CDN_BUCKET_NAME_LOGO,
            str(picture_id))

    def get_place(
            self,
            place_id,
            account_id=None,
            check_status=True,
            connection=None,
            include_address=False,
            include_contacts=False,
            include_geometries=False,
            include_photos=False,
            locale=DEFAULT_LOCALE):
        """
        Return the information about the specified place.


        :param place_id: Identification of the place to return information.

        :param account_id: Identification of the account of behalf of whom
            the function is called.  If defined, the function returns
            customized information that this user may have defined for this
            place.

        :param check_status: Indicate whether the function must check the
            current status of this place and raise an exception if it is not
            of enabled.

        :param connection: A `RdbmsConnection` instance to be used supporting
            the Python clause `with ...:`.

        :param include_address: Indicate whether to return the address
            information of this place.

        :param include_contacts: Indicate whether to return the contact
            information of this place.

        :param include_geometries: Indicate whether to return spatial data
            (geometries) related to this place, such as its boundaries and its
            pathway.

        :param include_photos: Indicate whether to return a first set of
            photos taken at this place.

        :param locale: An object `Locale` to return textual information of the
            place.


        :return: An instance containing the following attributes:

            * `account_id`: Identification of the account of the user who
              registered this place.

            * `agent_id`: Identification of the account of the official
              representative of this place, i.e., the user user who claimed
              this place.

            * `category`: The category qualifying the best this place.

            * `cover_photo_id`: Identification of the cover photo of the place,
              if any defined.

            * `cover_photo_url`: Uniform Resource Locator (URL) that specifies
              the location of the cover photo of the place, if any defined.

            * `place_id`: Identification of the place.

            * `team_id`: Identification of the organization that manages
              this place.

            * `timezone`: time zone of the location of the place.  It is the
              difference between the time at this location and UTC (Universal
              Time Coordinated). UTC is also  known as GMT or  Greenwich Mean
              Time or Zulu Time.

            * `object_status`: Current status of this place.

            * `creation_time`: Time when this place has been registered
              against the platform.

            * `update_time`: Time of the most recent modification of one or
              more attributes of this place.


        :raise DeletedObjectException: if the place has been deleted while the
            argument `check_status` has been set to `True`.

        :raise DisabledObjectException: if the place has been disabled while
            the argument `check_status` has been set to `True`.

        :raise UndefinedObjectException: if the specified identification
            doesn't refer to a place registered against the platform.
        """
        with self.acquire_rdbms_connection(auto_commit=False, connection=connection) as connection:
            cursor = connection.execute(
                """
                SELECT 
                    account_id,
                    accuracy,
                    agent_id,
                    ST_Z(location) AS altitude,
                    area_id,
                    ST_AsText(boundaries) AS boundaries,
                    cover_photo_id,
                    creation_time,
                    is_address_edited,
                    is_location_edited,
                    ST_Y(location) AS latitude,
                    ST_X(location) AS longitude,
                    object_status,
                    ST_AsText(pathway) AS pathway,
                    picture_id,
                    place_id,
                    radius,
                    reviewer_id,
                    team_id,
                    update_time,
                    visibility
                  FROM 
                    place
                  WHERE 
                    place_id = %(place_id)s
                    AND (object_status = %(OBJECT_STATUS_ENABLED)s
                         OR (object_status IN (%(OBJECT_STATUS_PENDING)s, %(OBJECT_STATUS_DISABLED)s)  
                             AND (account_id = %(account_id)s
                                  OR (team_id IS NOT NULL AND team_is_administrator(account_id, team_id)))))
                    AND (visibility = %(VISIBILITY_PUBLIC)s 
                         OR (visibility = %(VISIBILITY_PRIVATE)s AND account_id = %(account_id)s)
                         OR (visibility = %(VISIBILITY_TEAM)s
                             AND (account_id = %(account_id)s OR team_is_member(account_id, team_id))))
                """,
                {
                    'OBJECT_STATUS_DISABLED': ObjectStatus.disabled,
                    'OBJECT_STATUS_ENABLED': ObjectStatus.enabled,
                    'OBJECT_STATUS_PENDING': ObjectStatus.pending,
                    'VISIBILITY_PRIVATE': Visibility.private,
                    'VISIBILITY_PUBLIC': Visibility.public,
                    'VISIBILITY_TEAM': Visibility.team,
                    'account_id': account_id,
                    'place_id': place_id
                })

            row = cursor.fetch_one()
            if row is None:
                raise self.UndefinedObjectException(
                    "The specified place is not registered to the platform",
                    payload={'place_id': place_id})

            # Build the instance containing the information of the place.
            place = row.get_object({
                'account_id': cast.string_to_uuid,
                'cover_photo_id': cast.string_to_uuid,
                'creation_time': cast.string_to_timestamp,
                'agent_id': cast.string_to_uuid,
                'team_id': cast.string_to_uuid,
                'update_time': cast.string_to_timestamp,
                'place_id': cast.string_to_uuid })

            # Check whether the place is currently enable, and raise an exception
            # if not.
            if check_status:
                if place.object_status == ObjectStatus.disabled:
                    raise self.DisabledObjectException()
                elif place.object_status == ObjectStatus.deleted:
                    raise self.DeletedObjectException()

            # Build the Uniform Resource Locator (URL) of the place's cover photo,
            # if any.
            place.cover_photo_url = place.cover_photo_id and os.path.join(
                settings.CDN_URL_HOSTNAME,
                self.CDN_BUCKET_NAME_PHOTO,
                str(place.cover_photo_id))

            # Build the geographical location of the place.
            if place.longitude and place.latitude:
                GeoPoint.objectify_attributes(place)

            if not include_geometries:
                del place.boundaries
                del place.pathway

            # Include the address information of the place in the required locale
            # if defined, or the default locale.
            if include_address:
                place.address = self.get_place_address(
                    place_id,
                    account_id=account_id,
                    connection=connection,
                    locale=locale)

            # Include the contacts information of the place, when required.
            if include_contacts:
                place.contacts = self.get_place_contacts(
                    place_id,
                    account_id=account_id,
                    connection=connection)

            # Include an initial set of photos of the place, when required.
            if include_photos:
                place.photos = self.__get_place_photos(
                    place_id,
                    account_id=account_id,
                    connection=connection)

            return place

    def get_places(
            self,
            place_ids,
            account_id=None,
            connection=None,
            include_address=False,
            include_contacts=False,
            include_geometries=False,
            include_photos=False,
            locale=DEFAULT_LOCALE):
        """
        Return extended information about the specified places.


        :param place_ids: A list of identifications of the places to return
            information.

        :param account_id: Identification of the account of behalf of whom
            the function is called.  If defined, the function returns
            customized information that this user may have defined for these
            places.

        :param connection: An object `RdbmsConnection` supporting the Python
            clause `with ...:`.

        :param include_address: Indicate whether to return the address
            information of each place.

        :param include_contacts: Indicate whether to return the contact
            information of each place.

        :param include_geometries: Indicate whether to return spatial data
            (geometries) related to this place, such as its boundaries and its
            pathway.

        :param include_photos: Indicate whether to return a list of photos
            related to each place.

        :param locale: An object `Locale` to return textual information of
            each place.

        :return:
        """
        with self.acquire_rdbms_connection(auto_commit=False, connection=connection) as connection:
            cursor = connection.execute(
                """
                SELECT
                    account_id,
                    accuracy,
                    agent_id,
                    ST_Z(location) AS altitude,
                    area_id,
                    ST_AsText(boundaries) AS boundaries,
                    cover_photo_id,
                    creation_time,
                    is_address_edited,
                    is_location_edited,
                    ST_Y(location) AS latitude,
                    ST_X(location) AS longitude,
                    object_status,
                    ST_AsText(pathway) AS pathway,
                    picture_id,
                    place_id,
                    radius,
                    reviewer_id,
                    team_id,
                    update_time,
                    visibility
                  FROM
                    place
                  WHERE
                    place_id IN (%(place_ids)s)
                """,
                {
                    'place_ids': place_ids
                })

            # @todo: convert "pathway" and "boundaries" into list of points.
            places = {
                place.place_id: place
                for place in [
                    row.get_object({
                        'account_id': cast.string_to_uuid,
                        'agent_id': cast.string_to_uuid,
                        'area_id': cast.string_to_uuid,
                        'cover_photo_id': cast.string_to_uuid,
                        'creation_time': cast.string_to_timestamp,
                        'object_status': ObjectStatus,
                        'picture_id': cast.string_to_uuid,
                        'place_id': cast.string_to_uuid,
                        'reviewer_id': cast.string_to_uuid,
                        'team_id': cast.string_to_uuid,
                        'update_time': cast.string_to_timestamp,
                        'visibility': Visibility
                    })
                    for row in cursor.fetch_all()
                ]
            }

            for place in places.values():
                if place.longitude and place.latitude:
                    GeoPoint.objectify_attributes(place)

                place.picture_url = place.picture_id and os.path.join(
                    settings.CDN_URL_HOSTNAME,
                    self.CDN_BUCKET_NAME_LOGO,
                    str(place.picture_id))

                place.cover_photo_url = place.cover_photo_id and os.path.join(
                    settings.CDN_URL_HOSTNAME,
                    self.CDN_BUCKET_NAME_PHOTO,
                    str(place.cover_photo_id))

                if not include_geometries:
                    del place.boundaries
                    del place.pathway

            # Retrieve the address information of the specified places, when
            # requested.
            if include_address:
                places_address = self.get_places_address(
                    place_ids,
                    account_id=account_id,
                    connection=connection,
                    locale=locale)

                for place_id, place_address in places_address.items():
                    places[place_id].address = place_address

            # Retrieve the contact information of the specified places, when
            # requested.
            if include_contacts:
                places_contacts = self.get_places_contacts(
                    place_ids,
                    account_id=account_id,
                    connection=connection)

                for place_id, place_contacts in places_contacts.items():
                    places[place_id].contacts = place_contacts

            return list(places.values())


































    # INSERT INTO place_index
    #     return place


    # def get_distance(self, place_id, location,
    #         connection=None):
    #     """
    #     Return the linear distance in meters between the place and the
    #     specified location.  The function uses a spherical earth and radius
    #     of 6370986 meters
    #
    #     If the place is defined a boundary, the function returns the linear
    #     distance between the closest point and the specified location,
    #     otherwise it returns the linear distance between the centroid location
    #     of the place and the specified location.
    #
    #
    #     :param place_id: identification of a place.
    #
    #     :param location: the geographical location to return the distance from
    #         the place.
    #
    #     :param connection: a `RdbmsConnection` instance to be used         supporting the Python clause `with ...:`.
    #
    #
    #     :return: the linear distance in meters between the place and the
    #         specified location.
    #     """
    #     with self.acquire_rdbms_connection(auto_commit=True, connection=connection) as connection:
    #         cursor = connection.execute("""
    #             SELECT ST_Distance_Sphere(
    #                      COALESCE(ST_3DClosestPoint(boundaries, ST_SetSRID(ST_MakePoint(%(longitude)s, %(latitude)s, %(altitude)s), 4326)),
    #                               location),
    #                      ST_SetSRID(ST_MakePoint(%(longitude)s, %(latitude)s, %(altitude)s), 4326)) AS distance
    #               FROM place
    #               WHERE place_id = %(place_id)s
    #                 AND object_status = %(OBJECT_STATUS_ENABLED)s""",
    #             {'OBJECT_STATUS_ENABLED': ObjectStatus.enabled,
    #              'altitude': location.altitude,
    #              'latitude': location.latitude,
    #              'longitude': location.longitude,
    #              'place_id': place_id})
    #
    #         row = cursor.fetch_one()
    #         if row is None:
    #             raise self.UndefinedObjectException('The specified identificartion does not refer to a place registered to the platform',
    #                     payload={ 'place_id': place_id })
    #
    #         return row.get_value('distance')




#     def search_places(
#             self,
#             app_id,
#             account_id=None,
#             area_id=None,
#             bounds=None,
#             categories=None,
#             intent=SearchIntent.browse,
#             include_address=False,
#             include_contacts=False,
#             include_photos=False,
#             ip_address=None,
#             keywords=None,
#             limit=None,
#             locale=DEFAULT_LOCALE,
#             location=None,
#             offset=0,
#             radius=DEFAULT_SEARCH_RADIUS,
#             sync_time=None,
#             team_id=None):
#         """
#         Return up to 10 places, worth of extended information, corresponding
#         to the specified criteria.
#
#
#         :param app_id: Identification of the client application such as a Web,
#             a desktop, or a mobile application, that accesses the service.
#
#         :param account_id: Identification of the account of the user of behalf
#             of whom this search is initiated.
#
#         :param bounds: A tuple of two instances `GeoPoint` that represent
#             the north-east corner and the south-west corners of the rectangle
#             area to search places in.
#
#         :param categories: A list of identifications of categories that
#             indicate the particular use that places to return have been
#             designed for.
#
#         :param include_address: Indicate whether to include the address
#             information of the places returned.
#
#         :param include_contacts: Indicate whether to include the contacts
#             information of the places returned.
#
#         :param include_photos: Indicate whether to include a limited set of
#             photos associated to the places that are returned.
#
#         :param ip_address: IPv4 address of the machine of the user, a tuple
#             consisting of four decimal numbers, each ranging from `0` to
#             `255`.  The IP address is used to geolocate the user.
#
#         :param intent: An item of the enumeration `SearchItent` that
#             indicates the intent in performing the search.
#
#         :param keywords: A list of keywords to search schools for.
#
#         :param limit: Constrain the number of places that are returned to the
#             specified number.
#
#         :param locale: An instance `Locale` to return textual information of
#             the places.
#
#         :param location: An instance `GeoPoint` of the geographic location
#             to search nearby places.
#
#         :param offset: Require to skip that many places before beginning to
#             return them.  If both `limit` and `offset` are specified, then
#             `offset` places are skipped before starting to count the `limit`
#             places that are returned.  The default value is `0`.
#
#         :param radius: Maximal distance in meters of the radius of search for
#             places nearby the specified location.
#
#
#         :return: A list of instances containing the following attributes:
#
#             * `address` (optional): Postal address of the place.  The address is
#               composed of one or more address components. The address components are
#               represented with a dictionary, where the key is as an item of the
#               enumeration `AddressComponentName` and the value is the textual
#               information of this component, written in the specified locale.
#
#             * `category` (required): code name of category qualifying this
#               place.
#
#             * `contacts` (optional): list of properties such as e-mail
#               addresses, phone numbers, etc., in respect of the electronic
#               business card specification (vCard).  The contact information is
#               represented by a list of tuples of the following form::
#
#                  (name:ContactName, value:string, is_primary:boolean, is_verified:boolean)
#
#               where:
#
#               * `name`: type of this contact information, which can be one
#                 of these standard names in respect of the electronic business
#                 card specification (vCard).
#
#               * `value`: value of this contact information representing by a
#                 string, such as `+84.01272170781`, the formatted value for a
#                 telephone number property.
#
#               * `is_primary`: indicate whether this contact information is
#                 the primary for this place.
#
#               * `is_verified`: indicate whether this contact information has
#                 been verified, whether it has been grabbed from a trusted
#                 Social Networking Service (SNS), or whether through a
#                 challenge/response process.
#
#             * `location` (required): geographic coordinates of the location of
#               the place represented with the following JSON structure::
#
#                   {
#                     "accuracy": decimal,
#                     "altitude": decimal,
#                     "latitude": decimal,
#                     "longitude": decimal
#                   }
#
#               where:
#
#               * `accuracy` (optional): accuracy of the place's position in
#                 meters.
#
#               * `altitude` (optional): altitude in meters of the place.
#
#               * `latitude` (required): latitude-angular distance, expressed
#                 in decimal degrees (WGS84 datum), measured from the center of the
#                 Earth, of a point north or south of the Equator corresponding to
#                 the place's location.
#
#               * `longitude` (required): longitude-angular distance,
#                 expressed in decimal degrees (WGS84 datum), measured from the
#                 center of the Earth, of a point east or west of the Prime
#                 Meridian corresponding to the place's location.
#
# * `picture_id` (optional): Identification of the logo picture representing
#   this place.
#
# * `picture_url` (optional): Uniform Resource Locator (URL) that specifies
#   the location of the logo of the place, if any defined.  The client
#   application can use this URL and append the query parameter `size` to
#   specify a given pixel resolution of this photo, such as `thumbnail`,
#   `small`, `medium`, `large` (cf. `settings.IMAGE_PIXEL_RESOLUTIONS['place']`).
#
#             * `place_id` (required): identification of the place.
#
#             * `score` (optional): estimation of the probability that the
#               place is relevant to the search, given the intent.  A perfect
#               precision score of `1.0` means that the place returned is
#               totally relevant.  This value is only returned if the intent of
#               the search is `checkin` or `social`.
#
#             * `update_time` (required): time of the last modification of one
#               or more attributes of this place.  This time should be used by
#               the client application to manage its cache of places and to
#               reduce the average time to access data of places.  When the
#               client application needs to read places' attributes, it first
#               checks whether a copy of these data is in its cache.  If so, the
#               client application immediately reads from the cache, which is
#               much faster than requesting these data from the server platform.
#
#
#         :raise InvalidArgumentException: if no location provider, such as the
#             arguments `area_id`, `location`, `bounds`, or `ip_address`,
#             has been passed to this function.
#
#
# :warning: boundaries is a list of `(longitude, latitude)`, while it
#     should also have `altitude`.
#         """
#         location_provider_count = sum([provider is not None for provider in [ area_id, location, bounds, ip_address]])
#         if keywords is None and location_provider_count != 1:
#             raise self.InvalidArgumentException('One and only one geographical location provider MUST be provided')
#
#         if bounds:
#             southwest_location, northeast_location = bounds
#
#             boundaries = [
#                 (southwest_location.longitude, northeast_location.latitude),
#                 (northeast_location.longitude, northeast_location.latitude),
#                 (northeast_location.longitude, southwest_location.latitude),
#                 (southwest_location.longitude, southwest_location.latitude),
#                 (southwest_location.longitude, northeast_location.latitude)
#             ]
#
#             geometry = 'LINESTRING({})'.format(','.join([
#                 f'{longitude} {latitude}' for (longitude, latitude) in boundaries]))
#
#         elif ip_address:
#             areas = AreaService().find_areas_with_ip_address(
#                 app_id,
#                 ip_address,
#                 lowest_area_level=0,
#                 highest_area_level=0,
#                 locale=locale)
#
#             if len(areas) == 0:
#                 raise self.InvalidOperationException('No geographical area has been found for the specified IP address')
#
#             area = areas[0]
#
#         # Find places that correspond to the specified criteria.
#         with self.acquire_rdbms_connection(False) as connection:
#             if keywords:
#                 # Cleanse and latinize the given keywords, contained in the specified
#                 # string.
#                 cleansed_keywords = string_util.string_to_keywords(keywords)
#
#                 cursor = connection.execute(
#                     """
#                     SELECT
#                         place_id,
#                         picture_id,
#                         category,
#                         category_id,
#                         ST_X(location) AS longitude,
#                         ST_Y(location) AS latitude,
#                         ST_Z(location) AS altitude,
#                         accuracy,
#                         radius,
#                         cover_photo_id,
#                         object_status,
#                         update_time,
#                         score
#                       FROM (
#                         SELECT
#                             place_id,
#                             COUNT(*) AS score
#                           FROM place
#                           INNER JOIN place_index
#                             USING (place_id)
#                           WHERE keyword ~ %(keywords)s
#                             AND (%(categories)s IS NULL OR category IN (%(categories)s))
#                             AND (object_status = %(OBJECT_STATUS_ENABLED)s
#                                  OR (object_status = %(OBJECT_STATUS_PENDING)s AND account_id = %(account_id)s))
#                             AND ((%(longitude)s IS NULL AND %(area_id)s IS NULL)
#                                  OR (%(longitude)s IS NOT NULL AND ST_Distance_Sphere(
#                                          location,
#                                          ST_SetSRID(ST_MakePoint(%(longitude)s, %(latitude)s), 4326)) <= %(radius)s)
#                                  OR (%(area_id)s IS NOT NULL AND ST_Contains(get_area_boundaries(%(area_id)s), location)))
#                           GROUP BY place_id) AS foo
#                       INNER JOIN place
#                         USING (place_id)
#                       ORDER BY
#                           score DESC,
#                           place_id DESC
#                       LIMIT %(limit)s
#                       OFFSET %(offset)s
#                     """,
#                     {
#                         'OBJECT_STATUS_ENABLED': ObjectStatus.enabled,
#                         'OBJECT_STATUS_PENDING': ObjectStatus.pending,
#                         'account_id': account_id,
#                         'area_id': area_id or (ip_address and area.area_id),
#                         'categories': categories,
#                         'keywords': keywords and '|'.join(cleansed_keywords),
#                         'latitude': location and location.latitude,
#                         'limit': min(limit, self.MAXIMUM_LIMIT) or self.MAXIMUM_LIMIT,
#                         'longitude': location and location.longitude,
#                         'offset': offset,
#                         'radius': min(radius, self.MAXIMAL_SEARCH_RADIUS)
#                     })
#             else:
#                 cursor = connection.execute(
#                     """
#                     SELECT
#                         place_id,
#                         picture_id,
#                         category,
#                         category_id,
#                         ST_X(location) AS longitude,
#                         ST_Y(location) AS latitude,
#                         ST_Z(location) AS altitude,
#                         accuracy,
#                         radius,
#                         cover_photo_id,
#                         object_status,
#                         update_time
#                       FROM place
#                       WHERE (%(categories)s IS NULL OR category IN (%(categories)s))
#                         AND (object_status = %(OBJECT_STATUS_ENABLED)s
#                              OR (object_status = %(OBJECT_STATUS_PENDING)s AND account_id = %(account_id)s))
#                         AND ((%(longitude)s IS NOT NULL AND ST_Distance_Sphere(location,
#                                 ST_SetSRID(ST_MakePoint(%(longitude)s, %(latitude)s), 4326)) <= %(radius)s)
#                              OR (%(area_id)s IS NOT NULL AND ST_Contains(get_area_boundaries(%(area_id)s), location)))
#                       LIMIT %(limit)s
#                     """,
#                     {
#                         'OBJECT_STATUS_ENABLED': ObjectStatus.enabled,
#                         'OBJECT_STATUS_PENDING': ObjectStatus.pending,
#                         'account_id': account_id,
#                         'area_id': area_id or (ip_address and area.area_id),
#                         'categories': categories,
#                         'latitude': location and location.latitude,
#                         'limit': min(limit, self.MAXIMUM_LIMIT) or self.MAXIMUM_LIMIT,
#                         'longitude': location and location.longitude,
#                         # 'offset': offset,
#                         'radius': min(radius, self.MAXIMAL_SEARCH_RADIUS)
#                     })
#
#             places = dict([
#                 (place.place_id, place)
#                 for place in [ row.get_object({
#                     'category_id': cast.string_to_uuid,
#                     'cover_photo_id': cast.string_to_uuid,
#                     'picture_id': cast.string_to_uuid,
#                     'place_id': cast.string_to_uuid,
#                     'update_time': cast.string_to_timestamp
#                 })
#                     for row in cursor.fetch_all()]])
#
#             if len(places):
#                 # Gather all the information related to the location of the place into
#                 # an object that is set as a specific attribute of the place.
#                 for place in places.values():
#                     if place.longitude and place.latitude:
#                         place.location = GeoPoint(
#                             place.latitude,
#                             place.longitude,
#                             altitude=place.altitude,
#                             accuracy=place.accuracy)
#                     del place.latitude, place.longitude, place.altitude, place.accuracy
#
#                     place.picture_url = place.picture_id and os.path.join(
#                         settings.CDN_URL_HOSTNAME,
#                         self.CDN_BUCKET_NAME_LOGO,
#                         str(place.picture_id))
#
#                     place.cover_photo_url = place.cover_photo_id and os.path.join(
#                         settings.CDN_URL_HOSTNAME,
#                         self.CDN_BUCKET_NAME_PHOTO,
#                         str(place.cover_photo_id))
#
#                 # Include the address information of the returned places in the given
#                 # locale or the default locale.
#                 if include_address:
#                     cursor = connection.execute(
#                         """
#                         SELECT *
#                           FROM (
#                             SELECT place_id,
#                                    account_id,
#                                    property_name,
#                                    property_value,
#                                    locale,
#                                    row_number() OVER (
#                                        PARTITION BY place_id,
#                                                     property_name
#                                        ORDER BY compare_locale(locale, %(locale)s) DESC,
#                                                 account_id NULLS LAST) AS position
#                                 FROM place_address
#                                 WHERE place_id IN (%[place_ids]s)
#                                   AND ((%(account_id)s IS NULL AND account_id IS NULL) OR
#                                        (%(account_id)s IS NOT NULL AND (account_id IS NULL OR account_id = %(account_id)s)))
#                                   AND (compare_locale(locale, %(locale)s) > 0 OR
#                                        compare_locale(locale, %(DEFAULT_LOCALE)s) > 0)) AS foo
#                             WHERE position = 1
#                         """,
#                         {
#                             'DEFAULT_LOCALE': DEFAULT_LOCALE,
#                             'account_id': account_id,
#                             'locale': locale,
#                             'place_ids': list(places.keys())
#                         })
#
#                     components = collections.defaultdict(dict)
#                     for component in [
#                             row.get_object({'place_id': cast.string_to_uuid})
#                             for row in cursor.fetch_all()]:
#                         components[component.place_id][component.property_name] = component.property_value
#
#                     for place in places.values():
#                         if components.get(place.place_id):
#                             place.address = Object(**components[place.place_id])
#
#                 # Include the contacts information of the returned places.
#                 # if include_contacts:
#                 #     cursor = connection.execute("""
#                 #         SELECT *
#                 #           FROM (
#                 #             SELECT place_id,
#                 #                    property_name,
#                 #                    property_value,
#                 #                    is_primary,
#                 #                    is_verified,
#                 #                    row_number() OVER (
#                 #                        PARTITION BY place_id, property_name
#                 #                        ORDER BY account_id NULLS LAST) AS position
#                 #                 FROM place_contact
#                 #                 WHERE place_id IN (%[place_ids]s)
#                 #                   AND ((%(account_id)s IS NULL AND account_id IS NULL) OR
#                 #                        (%(account_id)s IS NOT NULL AND (account_id IS NULL OR account_id = %(account_id)s)))) AS foo
#                 #             WHERE position = 1""",
#                 #         { 'account_id': account_id,
#                 #           'place_ids': places.keys() })
#                 #
#                 #     for place in places.values():
#                 #         place.contacts = []
#                 #
#                 #     for contact in [ row.get_object({ 'place_id': cast.string_to_uuid })
#                 #             for row in cursor.fetch_all() ]:
#                 #         places[contact.place_id].contacts.append(
#                 #                 (contact.property_name, contact.property_value,
#                 #                  contact.is_primary, contact.is_verified) )
#
#             # if intent == SearchIntent.checkin:
#             #     connection.execute("""
#             #         INSERT INTO _place_search_history_(
#             #                         imei,
#             #                         device_model,
#             #                         os_version,
#             #                         location,
#             #                         accuracy,
#             #                         place_id)
#             #           VALUES (%(imei)s,
#             #                   %(device_model)s,
#             #                   %(os_version)s,
#             #                   ST_SetSRID(ST_MakePoint(%(longitude)s, %(latitude)s, %(altitude)s), 4326),
#             #                   %(accuracy)s,
#             #                   %(place_id)s)""",
#             #         { 'accuracy': location.accuracy,
#             #           'altitude': location.altitude or 0,
#             #           'device_model': _device_model_,
#             #           'imei': _imei_,
#             #           'latitude': location.latitude,
#             #           'longitude': location.longitude,
#             #           'os_version': _os_version_,
#             #           'place_id': places.values()[0].place_id if len(places.values()) > 0 else None })
#
#             return list(places.values())
#
#
#     # def sync_place(self, app_id, account_id, place,
#     #         optimistic_match=False,
#     #         team_id=None,
#     #         visibility=Visibility.public):
#     #     with self.acquire_rdbms_connection(True) as connection:
#     #         #     cursor = connection.execute("""
#     #         #         SELECT place_id
#     #         #           FROM place
#     #         #           WHERE (boundaries IS NOT NULL AND %(boundaries_defined)s
#     #         #                  AND ST_Intersects(boundaries, ST_MakePolygon(ST_GeomFromText('LINESTRING(%s)', 4326))""",
#     #         #
#     #         #         { 'boundaries_defined': place.boundaries is not None,
#     #         #           (True, ','.join([ '%s %s %s' % (longitude, latitude, altitude)
#     #         #                         for (longitude, latitude, altitude) in place.boundaries ])),
#     #
#     #
#     #         cursor = connection.execute("""
#     #             INSERT INTO place(
#     #                     account_id,
#     #                     team_id,
#     #                     location,
#     #                     boundaries)
#     #               VALUES (%(account_id)s,
#     #                       %(team_id)s,
#     #                       ST_SetSRID(ST_MakePoint(%(longitude)s, %(latitude)s, %(altitude)s), 4326),
#     #                       ST_MakePolygon(ST_GeomFromText(%(boundaries)s, 4326)))
#     #               RETURNING place_id,
#     #                         update_time""",
#     #             { 'account_id': account_id,
#     #               'altitude': place.location and place.location.altitude,
#     #               'boundaries': 'LINESTRING(%s)' % ','.join([ '%s %s %s' % (longitude, latitude, altitude)
#     #                     for (longitude, latitude, altitude) in place.boundaries ]),
#     #               'latitude': place.location and place.location.latitude,
#     #               'longitude': place.location and place.location.longitude,
#     #               'team_id': team_id })
#     #
#     #         result = cursor.fetch_one().get_object({
#     #                 'place_id': cast.string_to_uuid,
#     #                 'update_time': cast.string_to_timestamp })
#     #
#     #         if place.address:
#     #             if isinstance(place.address, dict):
#     #                 values = [ (result.place_id, account_id, locale, property_name, property_value)
#     #                         for property_name, property_value in place.address.iteritems() ]
#     #             else:
#     #                 values = [ (result.place_id, account_id, address[AddressComponentName.locale], property_name, property_value)
#     #                         for address in place.address
#     #                             for property_name, property_value in address.iteritems()
#     #                                 if property_name != AddressComponentName.locale ]
#     #
#     #             connection.execute("""
#     #                 INSERT INTO place_address(
#     #                         place_id,
#     #                         account_id,
#     #                         locale,
#     #                         property_name,
#     #                         property_value)
#     #                   VALUES %[values]s""",
#     #                { 'values': values })
#     #
#     #         if place.contacts:
#     #             connection.execute("""
#     #                 INSERT INTO place_contact(
#     #                         place_id,
#     #                         property_name,
#     #                         property_value,
#     #                         is_primary,
#     #                         is_verified)
#     #                   VALUES %[values]s""",
#     #                 { 'values': [ (result.place_id, property_type, property_value, is_primary, is_verified)
#     #                         for (property_type, property_value, is_primary, is_verified) in place.contacts ] })
